import logging
from dataclasses import (
    dataclass,
)

from fa_purity import (
    Cmd,
)
from fa_purity.json import (
    JsonObj,
)

from . import (
    _access,
    _refresh,
    _revoke,
)
from ._core import (
    Credentials,
    RefreshToken,
    Token,
)

ACCOUNTS_URL = "https://accounts.zoho.com"  # for US region
LOG = logging.getLogger(__name__)


@dataclass(frozen=True)
class AuthApi:
    new_access_token: Cmd[Token]
    manual_new_refresh_token: Cmd[RefreshToken]
    manual_revoke_token: Cmd[JsonObj]


@dataclass(frozen=True)
class AuthApiFactory:
    @staticmethod
    def auth_api(creds: Credentials) -> AuthApi:
        return AuthApi(
            _access.new_access_token(creds),
            _refresh.generate_refresh_token(creds),
            _revoke.revoke_refresh_token(),
        )


__version__ = "1.0.0"
__all__ = [
    "Credentials",
    "RefreshToken",
    "Token",
]



from arch_lint.dag.check import (
    dag_map_completeness,
)
from arch_lint.forbidden import (
    check_forbidden,
)
from arch_lint.graph import (
    ImportGraph,
)
from arch_lint.private import (
    check_private,
)

from .arch import (
    forbidden_allowlist,
    project_dag,
)

root = "fluidattacks_zoho_sdk"


def test_dag_creation() -> None:
    project_dag()


def test_dag_completeness() -> None:
    graph = ImportGraph.build_graph(root, True)
    dag_map_completeness(project_dag(), graph, next(iter(graph.roots)))


def test_forbidden_creation() -> None:
    forbidden_allowlist()


def test_forbidden() -> None:
    graph = ImportGraph.build_graph(root, True)
    allowlist_map = forbidden_allowlist()
    check_forbidden(allowlist_map, graph)


def test_private() -> None:
    graph = ImportGraph.build_graph(root, False)
    check_private(graph, next(iter(graph.roots)))



from __future__ import (
    annotations,
)

import logging
from collections.abc import Callable
from dataclasses import (
    dataclass,
)
from typing import (
    TypeVar,
)

from fa_purity import Cmd, Coproduct, FrozenDict, FrozenList, Maybe, Result
from fa_purity.json import JsonObj, JsonPrimitiveFactory, JsonUnfolder, JsonValue
from fluidattacks_etl_utils.decode import int_to_str
from fluidattacks_etl_utils.smash import bind_chain
from pure_requests import (
    response,
)
from pure_requests import (
    retry as _retry,
)
from pure_requests.basic import Data, Endpoint, HttpClient, HttpClientFactory, Params
from pure_requests.retry import (
    HandledError,
    MaxRetriesReached,
)
from requests import Response

from fluidattacks_zoho_sdk.auth import AuthApiFactory, Credentials, Token
from fluidattacks_zoho_sdk.ids import OrgId

from ._core import (
    HandledErrors,
    HTTPError,
    HttpJsonClient,
    JSONDecodeError,
    RelativeEndpoint,
    RequestException,
    UnhandledErrors,
)

LOG = logging.getLogger(__name__)

_S = TypeVar("_S")
_F = TypeVar("_F")
HTTP_UNAUTHORIZED: int = 401


def _refresh_access_token(creds: Credentials) -> Cmd[Token]:
    return AuthApiFactory.auth_api(creds).new_access_token


def _is_404(error: Maybe[HTTPError]) -> tuple[bool, int]:
    status: int = error.map(
        lambda v: v.raw.response.status_code,  # type: ignore[misc]
    ).value_or(0)
    status_is_404: bool = (
        error.map(lambda e: e.raw.response.status_code == HTTP_UNAUTHORIZED).value_or(False)  # type: ignore[misc]
    )
    return (status_is_404, status)


def _extract_http_error(error: HandledError[HandledErrors, UnhandledErrors]) -> Maybe[HTTPError]:
    return error.value.map(
        lambda handled: handled.error.map(
            lambda http_err: Maybe.some(http_err),
            lambda _: Maybe.empty(),
        ),
        lambda _: Maybe.empty(),
    )


def _retry_cmd(
    retry: int,
    item: Result[_S, _F],
    client_zoho: Client,
    make_request: Callable[[HttpClient], Cmd[Result[_S, _F]]],
) -> Cmd[Result[_S, _F]]:
    def handle_401(
        error: _F,
    ) -> Cmd[Result[_S, _F]]:
        is_401 = _is_404(_extract_http_error(error))  # type: ignore[arg-type]
        if is_401[0]:
            return Cmd.wrap_impure(lambda: LOG.info("Refreshing token...")) + _refresh_access_token(
                client_zoho.creds,
            ).bind(
                lambda acess_token: make_request(
                    HttpClientFactory.new_client(
                        None,
                        client_zoho.build_headers(acess_token),
                        False,
                    ),
                ),
            )
        log = Cmd.wrap_impure(
            lambda: LOG.info(
                "retry #%2s waiting... ",
                retry,
            ),
        )
        return _retry.cmd_if_fail(item, log + _retry.sleep_cmd(retry**2))

    return item.map(lambda _: Cmd.wrap_value(item)).alt(handle_401).to_union()


def _http_error_handler(
    error: HTTPError,
) -> HandledError[HandledErrors, UnhandledErrors]:
    err_code: int = error.raw.response.status_code  # type: ignore[misc]
    handled = (
        401,
        429,
    )
    if err_code in range(500, 600) or err_code in handled:
        return HandledError.handled(HandledErrors(Coproduct.inl(error)))
    return HandledError.unhandled(UnhandledErrors(Coproduct.inr(Coproduct.inl(error))))


def _handled_request_exception(
    error: RequestException,
) -> HandledError[HandledErrors, UnhandledErrors]:
    return (
        error.to_chunk_error()
        .map(lambda e: HandledError.handled(HandledErrors(Coproduct.inr(Coproduct.inl(e)))))
        .lash(
            lambda _: error.to_connection_error().map(
                lambda e: HandledError.handled(HandledErrors(Coproduct.inr(Coproduct.inr(e)))),
            ),
        )
        .value_or(HandledError.unhandled(UnhandledErrors(Coproduct.inr(Coproduct.inr(error)))))
    )


def _handled_errors(
    error: Coproduct[JSONDecodeError, Coproduct[HTTPError, RequestException]],
) -> HandledError[HandledErrors, UnhandledErrors]:
    """Classify errors."""
    return error.map(
        lambda _: HandledError.unhandled(UnhandledErrors(error)),
        lambda c: c.map(
            _http_error_handler,
            _handled_request_exception,
        ),
    )


def _handled_errors_response(
    error: Coproduct[HTTPError, RequestException],
) -> HandledError[HandledErrors, UnhandledErrors]:
    return error.map(
        _http_error_handler,
        _handled_request_exception,
    )


def _adjust_unhandled(
    error: UnhandledErrors | MaxRetriesReached,
) -> Coproduct[UnhandledErrors, MaxRetriesReached]:
    return Coproduct.inr(error) if isinstance(error, MaxRetriesReached) else Coproduct.inl(error)


@dataclass(frozen=True)
class Client:
    _creds: Credentials
    _max_retries: int
    _org_id: OrgId

    def _full_endpoint(self, endpoint: RelativeEndpoint) -> Endpoint:
        return Endpoint("/".join(("https://desk.zoho.com/api/v1", *endpoint.paths)))

    @staticmethod
    def new(creds: Credentials, org_id: OrgId) -> Client:
        return Client(creds, 3, org_id)

    def _headers(self, acess_token: Token) -> JsonObj:
        return FrozenDict(
            {
                "orgId": JsonValue.from_primitive(
                    JsonPrimitiveFactory.from_raw(int_to_str(self._org_id.org_id.value)),
                ),
                "Authorization": JsonValue.from_primitive(
                    JsonPrimitiveFactory.from_raw(f"Zoho-oauthtoken {acess_token.raw_token}"),
                ),
            },
        )

    def get(
        self,
        endpoint: RelativeEndpoint,
        acess_token: Token,
        params: JsonObj,
    ) -> Cmd[
        Result[
            Coproduct[JsonObj, FrozenList[JsonObj]],
            Coproduct[UnhandledErrors, MaxRetriesReached],
        ]
    ]:
        _full = self._full_endpoint(endpoint)
        log = Cmd.wrap_impure(
            lambda: LOG.info("[API] get: %s\nparams = %s", _full, JsonUnfolder.dumps(params)),
        )
        client = HttpClientFactory.new_client(None, self._headers(acess_token), False)

        def make_request(
            new_client: HttpClient,
        ) -> Cmd[
            Result[
                Coproduct[JsonObj, FrozenList[JsonObj]],
                HandledError[HandledErrors, UnhandledErrors],
            ]
        ]:
            return (
                new_client.get(_full, Params(params))
                .map(
                    lambda r: r.alt(RequestException),
                )
                .map(
                    lambda r: bind_chain(r, lambda i: response.handle_status(i).alt(HTTPError)),
                )
                .map(
                    lambda r: bind_chain(
                        r,
                        lambda i: response.json_decode(i).alt(JSONDecodeError),
                    ).alt(
                        _handled_errors,
                    ),
                )
            )

        handled = log + make_request(client)
        return _retry.retry_cmd(
            handled,
            lambda retry, item: _retry_cmd(retry, item, self, make_request),
            self._max_retries,
        ).map(lambda r: r.alt(_adjust_unhandled))

    def post(
        self,
        endpoint: RelativeEndpoint,
        token: Token,
        params: JsonObj,
    ) -> Cmd[
        Result[
            Coproduct[JsonObj, FrozenList[JsonObj]],
            Coproduct[UnhandledErrors, MaxRetriesReached],
        ]
    ]:
        _full = self._full_endpoint(endpoint)
        log = Cmd.wrap_impure(lambda: LOG.info("[API] call (post): %s", _full))
        client = HttpClientFactory.new_client(None, self._headers(token), False)

        def make_request(
            new_client: HttpClient,
        ) -> Cmd[
            Result[
                Coproduct[JsonObj, FrozenList[JsonObj]],
                HandledError[HandledErrors, UnhandledErrors],
            ]
        ]:
            return (
                new_client.post(
                    self._full_endpoint(endpoint),
                    Params(FrozenDict({})),
                    Data(params),
                )
                .map(
                    lambda r: r.alt(RequestException),
                )
                .map(
                    lambda r: bind_chain(r, lambda i: response.handle_status(i).alt(HTTPError)),
                )
                .map(
                    lambda r: bind_chain(
                        r,
                        lambda i: response.json_decode(i).alt(JSONDecodeError),
                    ).alt(
                        _handled_errors,
                    ),
                )
            )

        handled = log + make_request(client)
        return _retry.retry_cmd(
            handled,
            lambda retry, item: _retry_cmd(retry, item, self, make_request),
            self._max_retries,
        ).map(lambda r: r.alt(_adjust_unhandled))

    def get_response(
        self,
        endpoint: RelativeEndpoint,
        token: Token,
        params: JsonObj,
    ) -> Cmd[Result[Response, Coproduct[UnhandledErrors, MaxRetriesReached]]]:
        _full = self._full_endpoint(endpoint)
        log = Cmd.wrap_impure(
            lambda: LOG.info("[API] get: %s\nparams = %s", _full, JsonUnfolder.dumps(params)),
        )
        client = HttpClientFactory.new_client(None, self._headers(token), False)

        def make_request(
            new_client: HttpClient,
        ) -> Cmd[
            Result[
                Response,
                HandledError[HandledErrors, UnhandledErrors],
            ]
        ]:
            return (
                new_client.get(_full, Params(params))
                .map(
                    lambda r: r.alt(RequestException),
                )
                .map(
                    lambda r: bind_chain(r, lambda i: response.handle_status(i).alt(HTTPError)),
                )
                .map(lambda r: r.alt(_handled_errors_response))
            )

        handled = log + make_request(client)
        return _retry.retry_cmd(
            handled,
            lambda retry, item: _retry_cmd(retry, item, self, make_request),
            self._max_retries,
        ).map(lambda r: r.alt(_adjust_unhandled))

    @property
    def client(self) -> HttpJsonClient:
        return HttpJsonClient(self.get, self.post, self.get_response)

    @property
    def creds(self) -> Credentials:
        return self._creds

    def build_headers(self, access_token: Token) -> JsonObj:
        return self._headers(access_token)



import inspect
import logging

from fa_purity import Cmd, FrozenList, Result, ResultE, Unsafe, cast_exception
from fa_purity.json import JsonObj, Primitive, UnfoldedFactory
from fluidattacks_etl_utils.bug import Bug
from fluidattacks_etl_utils.retry import retry_cmd
from requests import Response

from fluidattacks_zoho_sdk._decoders import assert_single
from fluidattacks_zoho_sdk._http_client import HttpJsonClient, RelativeEndpoint
from fluidattacks_zoho_sdk.auth import Token

from ._decode import decode_bulk_info, transform_columns
from .core import (
    BulkData,
    BulkEndpoint,
    BulkExportObj,
    BulkStatus,
    FileName,
    ModuleName,
    ShouldRetry,
)
from .utils import _handle_status, _wait_job, unzip_data

LOG = logging.getLogger(__name__)


def _retry_waiting_step(
    retry: int,
    result: ResultE[BulkStatus],
    job: BulkExportObj,
    token: Token,
    client: HttpJsonClient,
) -> Cmd[ResultE[BulkStatus]]:
    error_or_ok = result.to_coproduct().map(lambda _s: False, lambda e: isinstance(e, ShouldRetry))
    if error_or_ok is True:
        return _wait_job(retry, job.id_bulk).bind(
            lambda _: get_status_bulk_export(client, token, job).map(
                lambda r: r.bind(lambda tup: _handle_status(tup[1])),
            ),
        )

    return Cmd.wrap_value(result)


def create_bulk_export(
    client: HttpJsonClient,
    access_token: Token,
    module: ModuleName,
    endpoint_bulk: BulkEndpoint,
) -> Cmd[ResultE[tuple[BulkExportObj, BulkStatus]]]:
    endpoint = RelativeEndpoint.new(endpoint_bulk.route)
    params: dict[str, Primitive] = {"module": module.module}

    return client.post(endpoint, access_token, UnfoldedFactory.from_dict(params)).map(
        lambda result: (
            result.alt(
                lambda e: cast_exception(
                    Bug.new("_create_bulk_export", inspect.currentframe(), e, ()),
                ),
            )
            .bind(assert_single)
            .bind(decode_bulk_info)
        ),
    )


def get_status_retry(
    client: HttpJsonClient,
    token: Token,
    bulk: BulkExportObj,
    max_attempts: int,
) -> Cmd[ResultE[BulkStatus]]:
    def get_attempt(
        token: Token,
        obj: BulkExportObj,
        client: HttpJsonClient,
    ) -> Cmd[ResultE[BulkStatus]]:
        return get_status_bulk_export(client, token, obj).map(
            lambda result: (
                result.alt(
                    lambda e: cast_exception(
                        Bug.new("_get_bulk_export_status", inspect.currentframe(), e, ()),
                    ),
                ).bind(lambda tup: _handle_status(tup[1]))
            ),
        )

    log = Cmd.wrap_impure(lambda: LOG.info("[API]: Starting read status of bulk export"))
    handled = log + get_attempt(token, bulk, client)

    return retry_cmd(
        handled,
        lambda retry, result: _retry_waiting_step(retry, result, bulk, token, client),
        max_attempts,
    ).map(lambda r: r.alt(cast_exception))


def get_status_bulk_export(
    client: HttpJsonClient,
    access_token: Token,
    bulk: BulkExportObj,
) -> Cmd[ResultE[tuple[BulkExportObj, BulkStatus]]]:
    endpoint = RelativeEndpoint.new("bulkExport", bulk.id_bulk.id_bulk)
    params: dict[str, Primitive] = {}

    return client.get(endpoint, access_token, UnfoldedFactory.from_dict(params)).map(
        lambda result: (
            result.alt(
                lambda e: cast_exception(
                    Bug.new("_get_bulk_export_status", inspect.currentframe(), e, ()),
                ),
            )
            .bind(assert_single)
            .bind(decode_bulk_info)
        ),
    )


def _unwrap_and_unzip(
    result: Result[Response, Exception],
    file_name: FileName,
) -> Cmd[ResultE[BulkData]]:
    return result.to_coproduct().map(
        lambda inl: unzip_data(inl, file_name),
        lambda ri: Cmd.wrap_value(Result.failure(ri)),
    )


def download_bulk(
    client: HttpJsonClient,
    access_token: Token,
    bulk: BulkExportObj,
    file_name: FileName,
) -> Cmd[ResultE[BulkData]]:
    endpoint = RelativeEndpoint.new("downloadBulkExportFile")
    params: dict[str, Primitive] = {"exportId": bulk.id_bulk.id_bulk}

    return (
        client.get_response(endpoint, access_token, UnfoldedFactory.from_dict(params))
        .map(
            lambda result: result.alt(
                lambda e: cast_exception(
                    Bug.new("_download_bulk_export", inspect.currentframe(), e, ()),
                ),
            ),
        )
        .bind(lambda v: _unwrap_and_unzip(v, file_name))
    )


def _wait_and_download(
    bulk_obj: BulkExportObj,
    client: HttpJsonClient,
    token: Token,
    file_name: FileName,
) -> Cmd[ResultE[BulkData]]:
    return (
        get_status_retry(client, token, bulk_obj, 10)
        .map(lambda r: r.map(lambda _: bulk_obj))
        .map(lambda r: r.alt(Unsafe.raise_exception).to_union())
        .bind(lambda bulk: download_bulk(client, token, bulk, file_name))
    )


def transform_csv_phase(csv: ResultE[BulkData]) -> ResultE[FrozenList[JsonObj]]:
    bulk_data = csv.alt(Unsafe.raise_exception).to_union()
    return transform_columns(bulk_data)


def fetch_bulk(
    client: HttpJsonClient,
    access_token: Token,
    module: ModuleName,
    endpoint_bulk: BulkEndpoint,
    file_name: FileName,
) -> Cmd[ResultE[FrozenList[JsonObj]]]:
    create_bulk: Cmd[ResultE[tuple[BulkExportObj, BulkStatus]]] = create_bulk_export(
        client,
        access_token,
        module,
        endpoint_bulk,
    )

    def on_created(
        result: ResultE[tuple[BulkExportObj, BulkStatus]],
    ) -> Cmd[ResultE[FrozenList[JsonObj]]]:
        return (
            result.map(lambda pair: pair[0])
            .to_coproduct()
            .map(
                lambda bulk_obj: _wait_and_download(bulk_obj, client, access_token, file_name).map(
                    lambda data: transform_csv_phase(data),
                ),
                lambda err: Cmd.wrap_value(Result.failure(err)),
            )
        )

    return create_bulk.bind(on_created)



from __future__ import annotations

from collections.abc import Callable
from dataclasses import dataclass
from enum import Enum
from typing import IO, Final, final

from fa_purity import Cmd, FrozenList, ResultE
from fa_purity.json import JsonObj
from fluidattacks_etl_utils.handle_errors import handle_value_error

from fluidattacks_zoho_sdk.auth import Token


@dataclass(frozen=True)
class BulkExportId:
    id_bulk: str


@dataclass(frozen=True)
class ViewId:
    id_bulk: str


@dataclass(frozen=True)
class ModuleName:
    module: str


@final
class BulkStatus(Enum):
    QUEUED = "QUEUED"
    INITIATED = "INITIATED"
    IN_PROGRESS = "IN_PROGRESS"
    COMPLETED = "COMPLETED"
    FAILED = "FAILED"
    EXPIRED = "EXPIRED"
    CANCELLED = "CANCELLED"

    @staticmethod
    def from_raw(raw: str) -> ResultE[BulkStatus]:
        return handle_value_error(lambda: BulkStatus(raw))


NonTerminalStatus: Final[tuple[BulkStatus, ...]] = (
    BulkStatus.QUEUED,
    BulkStatus.IN_PROGRESS,
    BulkStatus.INITIATED,
)
TerminalStatus: Final[tuple[BulkStatus, ...]] = (
    BulkStatus.CANCELLED,
    BulkStatus.FAILED,
    BulkStatus.EXPIRED,
)


@dataclass(frozen=True)
class BulkExportObj:
    id_bulk: BulkExportId
    module: str


@dataclass(frozen=True)
class BulkEndpoint:
    route: str


@dataclass(frozen=True)
class ShouldRetry(Exception):
    status: BulkStatus


@dataclass(frozen=True)
class MustRestart(Exception):
    status: BulkStatus


@dataclass(frozen=True)
class BulkClient:
    create_bulk: Callable[
        [Token, ModuleName, BulkEndpoint],
        Cmd[ResultE[tuple[BulkExportObj, BulkStatus]]],
    ]
    get_status: Callable[[Token, BulkExportObj], Cmd[ResultE[tuple[BulkExportObj, BulkStatus]]]]
    download_bulk: Callable[[Token, BulkExportObj, FileName], Cmd[ResultE[BulkData]]]
    fetch_bulk: Callable[
        [Token, ModuleName, BulkEndpoint, FileName],
        Cmd[ResultE[FrozenList[JsonObj]]],
    ]


@dataclass(frozen=True)
class BulkData:
    file: IO[str]


@dataclass(frozen=True)
class FileName:
    name: str



from fa_purity import FrozenList, ResultE
from fa_purity.json import (
    JsonObj,
    JsonUnfolder,
    Unfolder,
)
from fluidattacks_etl_utils import smash
from fluidattacks_etl_utils.decode import DecodeUtils
from fluidattacks_etl_utils.natural import Natural

from fluidattacks_zoho_sdk._decoders import (
    decode_account_id,
    decode_account_id_ticket,
    decode_contact_id,
    decode_crm_id,
    decode_deparment_id,
    decode_id_team,
    decode_list_objs,
    decode_maybe_str,
    decode_optional_date,
    decode_product_id,
    decode_require_date,
    decode_team_id,
    decode_ticket_id,
    decode_user_id,
    decode_user_id_bulk,
)
from fluidattacks_zoho_sdk.ids import UserId
from fluidattacks_zoho_sdk.zoho_desk.core import (
    AgentObj,
    ContactAddres,
    ContactDates,
    ContactInfo,
    ContactObj,
    TeamObj,
    TicketDates,
    TicketObj,
    TicketProperties,
    UserObj,
)


def parse_bool(v: str) -> bool:
    lv = v.lower().strip()
    if lv == "true":
        return True
    if lv == "false":
        return False
    msg = f"Invalid boolean string: {v}"
    raise ValueError(msg)


def decode_user(raw: JsonObj) -> ResultE[UserObj]:
    return smash.smash_result_3(
        decode_user_id(raw),
        JsonUnfolder.optional(raw, "firstName", DecodeUtils.to_opt_str).map(
            lambda v: v.bind(lambda j: j),
        ),
        JsonUnfolder.require(raw, "lastName", DecodeUtils.to_str),
    ).map(lambda obj: UserObj(*obj))


def decode_user_bulk(raw: JsonObj) -> ResultE[UserObj]:
    return smash.smash_result_3(
        decode_user_id_bulk(raw),
        JsonUnfolder.optional(raw, "First Name", DecodeUtils.to_opt_str).map(
            lambda v: v.bind(lambda j: j),
        ),
        JsonUnfolder.require(raw, "Last Name", DecodeUtils.to_str),
    ).map(lambda v: UserObj(*v))


def decode_contact_addres(raw: JsonObj) -> ResultE[ContactAddres]:
    return smash.smash_result_3(
        JsonUnfolder.optional(raw, "city", DecodeUtils.to_opt_str).map(
            lambda v: v.bind(lambda x: x),
        ),
        JsonUnfolder.optional(raw, "country", DecodeUtils.to_opt_str).map(
            lambda v: v.bind(lambda x: x),
        ),
        JsonUnfolder.optional(raw, "street", DecodeUtils.to_opt_str).map(
            lambda v: v.bind(lambda x: x),
        ),
    ).map(lambda obj: ContactAddres(*obj))


def decode_contact_dates(raw: JsonObj) -> ResultE[ContactDates]:
    return smash.smash_result_2(
        JsonUnfolder.require(raw, "createdTime", DecodeUtils.to_date_time),
        JsonUnfolder.require(raw, "modifiedTime", DecodeUtils.to_date_time),
    ).map(lambda obj: ContactDates(*obj))


def decode_contact_info(raw: JsonObj) -> ResultE[ContactInfo]:
    return smash.smash_result_5(
        JsonUnfolder.require(raw, "email", DecodeUtils.to_str),
        JsonUnfolder.optional(raw, "facebook", DecodeUtils.to_opt_str).map(
            lambda v: v.bind(lambda x: x),
        ),
        JsonUnfolder.require(raw, "phone", DecodeUtils.to_str),
        JsonUnfolder.optional(raw, "mobile", DecodeUtils.to_opt_str).map(
            lambda v: v.bind(lambda x: x),
        ),
        JsonUnfolder.optional(raw, "secondaryEmail", DecodeUtils.to_opt_str).map(
            lambda v: v.bind(lambda x: x),
        ),
    ).map(lambda obj: ContactInfo(*obj))


def decode_contact_obj(raw: JsonObj) -> ResultE[ContactObj]:
    first = smash.smash_result_5(
        decode_user(raw),
        decode_contact_info(raw),
        decode_contact_addres(raw),
        decode_contact_dates(raw),
        decode_account_id(raw),
    )

    second = smash.smash_result_4(
        JsonUnfolder.require(raw, "ownerId", DecodeUtils.to_str)
        .bind(lambda v: Natural.from_int(int(v)))
        .map(lambda j: UserId(j)),
        decode_crm_id(raw),
        JsonUnfolder.optional(raw, "description", DecodeUtils.to_opt_str).map(
            lambda v: v.bind(lambda j: j),
        ),
        JsonUnfolder.optional(raw, "state", DecodeUtils.to_opt_str).map(
            lambda v: v.bind(lambda j: j),
        ),
    )

    three = smash.smash_result_3(
        JsonUnfolder.optional(raw, "title", DecodeUtils.to_opt_str).map(
            lambda v: v.bind(lambda j: j),
        ),
        JsonUnfolder.optional(raw, "type", DecodeUtils.to_opt_str).map(
            lambda v: v.bind(lambda j: j),
        ),
        JsonUnfolder.optional(raw, "zip", DecodeUtils.to_opt_str).map(
            lambda v: v.bind(lambda j: j),
        ),
    )

    return smash.smash_result_3(first, second, three).map(lambda v: ContactObj(*v[0], *v[1], *v[2]))


def decode_contacts(raw: JsonObj) -> ResultE[FrozenList[ContactObj]]:
    return JsonUnfolder.require(
        raw,
        "data",
        lambda v: Unfolder.to_list_of(v, lambda j: Unfolder.to_json(j).bind(decode_contact_obj)),
    )


def decode_ticket_dates(raw: JsonObj) -> ResultE[TicketDates]:
    return smash.smash_result_4(
        decode_optional_date(raw, "modifiedTime"),
        decode_require_date(raw, "createdTime"),
        decode_optional_date(raw, "closedTime"),
        decode_optional_date(raw, "customerResponseTime"),
    ).map(lambda v: TicketDates(*v))


def decode_ticket_properties(raw: JsonObj) -> ResultE[TicketProperties]:
    first = smash.smash_result_5(
        decode_ticket_id(raw),
        JsonUnfolder.require(raw, "ticketNumber", DecodeUtils.to_str),
        decode_maybe_str(raw, "subject"),
        decode_maybe_str(raw, "channel"),
        decode_maybe_str(raw, "status"),
    )

    second = smash.smash_result_4(
        decode_maybe_str(raw, "category"),
        JsonUnfolder.require(raw, "isEscalated", DecodeUtils.to_bool),
        JsonUnfolder.require(raw, "priority", DecodeUtils.to_str),
        decode_maybe_str(raw, "resolution"),
    )

    three = smash.smash_result_2(
        decode_maybe_str(raw, "clasification"),
        decode_maybe_str(raw, "description"),
    )
    return smash.smash_result_3(first, second, three).map(
        lambda v: TicketProperties(*v[0], *v[1], *v[2]),
    )


def decode_ticket_obj(raw: JsonObj) -> ResultE[TicketObj]:
    first = smash.smash_result_5(
        decode_account_id_ticket(raw),
        decode_deparment_id(raw),
        decode_team_id(raw),
        decode_product_id(raw),
        decode_ticket_dates(raw),
    )
    second = smash.smash_result_5(
        decode_ticket_properties(raw),
        decode_contact_id(raw),
        JsonUnfolder.require(raw, "email", DecodeUtils.to_str),
        JsonUnfolder.require(raw, "phone", DecodeUtils.to_str),
        decode_maybe_str(raw, "onholdTime"),
    )
    return smash.smash_result_2(first, second).map(lambda v: TicketObj(*v[0], *v[1]))


def decode_tickets(raw: JsonObj) -> ResultE[FrozenList[TicketObj]]:
    return JsonUnfolder.require(
        raw,
        "data",
        lambda v: Unfolder.to_list_of(v, lambda j: Unfolder.to_json(j).bind(decode_ticket_obj)),
    )


def decode_team_obj(raw: JsonObj) -> ResultE[TeamObj]:
    return smash.smash_result_4(
        decode_deparment_id(raw),
        decode_id_team(raw),
        JsonUnfolder.require(raw, "description", DecodeUtils.to_str),
        JsonUnfolder.require(raw, "name", DecodeUtils.to_str),
    ).map(lambda team: TeamObj(*team))


def decode_teams(raw: JsonObj) -> ResultE[FrozenList[TeamObj]]:
    return JsonUnfolder.require(
        raw,
        "teams",
        lambda v: Unfolder.to_list_of(v, lambda j: Unfolder.to_json(j).bind(decode_team_obj)),
    )


# DECODERS AGENTS


def decode_agent_obj(raw: JsonObj) -> ResultE[AgentObj]:
    first = smash.smash_result_3(
        decode_user_bulk(raw),
        JsonUnfolder.require(raw, "IsConfirmed", DecodeUtils.to_str).map(parse_bool),
        JsonUnfolder.require(raw, "Email", DecodeUtils.to_str),
    )

    second = smash.smash_result_3(
        JsonUnfolder.require(raw, "Profile", DecodeUtils.to_str),
        JsonUnfolder.require(raw, "Role", DecodeUtils.to_str),
        JsonUnfolder.require(raw, "Status", DecodeUtils.to_str),
    )
    return smash.smash_result_2(first, second).map(lambda v: AgentObj(*v[0], *v[1]))


def decode_agents(raws: FrozenList[JsonObj]) -> ResultE[FrozenList[AgentObj]]:
    return decode_list_objs(raws, decode_agent_obj)



from __future__ import (
    annotations,
)

from collections.abc import Callable
from dataclasses import (
    dataclass,
    field,
)

from fa_purity import Cmd, Coproduct, FrozenList, Result, ResultE
from fa_purity.json import (
    JsonObj,
)
from pure_requests.retry import (
    MaxRetriesReached,
)
from requests import Response
from requests.exceptions import (
    ChunkedEncodingError as RawChunkedEncodingError,
)
from requests.exceptions import (
    ConnectionError as RawConnectionError,
)
from requests.exceptions import (
    HTTPError as RawHTTPError,
)
from requests.exceptions import (
    JSONDecodeError as RawJSONDecodeError,
)
from requests.exceptions import (
    RequestException as RawRequestException,
)

from fluidattacks_zoho_sdk.auth import Token


@dataclass(frozen=True)
class UnhandledErrors:
    error: Coproduct[JSONDecodeError, Coproduct[HTTPError, RequestException]]


@dataclass(frozen=True)
class HandledErrors:
    error: Coproduct[
        HTTPError,
        Coproduct[ChunkedEncodingError, RequestsConnectionError],
    ]


@dataclass(frozen=True)
class _Private:
    pass


@dataclass(frozen=True)
class HTTPError:
    raw: RawHTTPError


@dataclass(frozen=True)
class JSONDecodeError:
    raw: RawJSONDecodeError


@dataclass(frozen=True)
class ChunkedEncodingError:
    _private: _Private = field(repr=False, hash=False, compare=False)
    raw: RawChunkedEncodingError


@dataclass(frozen=True)
class RequestsConnectionError:
    _private: _Private = field(repr=False, hash=False, compare=False)
    raw: RawConnectionError


@dataclass(frozen=True)
class RequestException:
    raw: RawRequestException

    def to_chunk_error(self) -> ResultE[ChunkedEncodingError]:
        if isinstance(self.raw, RawChunkedEncodingError):
            return Result.success(ChunkedEncodingError(_Private(), self.raw))
        return Result.failure(ValueError("Not a ChunkedEncodingError"))

    def to_connection_error(self) -> ResultE[RequestsConnectionError]:
        if isinstance(self.raw, RawConnectionError):
            return Result.success(RequestsConnectionError(_Private(), self.raw))
        return Result.failure(ValueError("Not a RequestsConnectionError"))


@dataclass(frozen=True)
class RelativeEndpoint:
    paths: FrozenList[str]

    @staticmethod
    def new(*args: str) -> RelativeEndpoint:
        return RelativeEndpoint(tuple(args))


@dataclass(frozen=True)
class HttpJsonClient:
    get: Callable[
        [RelativeEndpoint, Token, JsonObj],
        Cmd[
            Result[
                Coproduct[JsonObj, FrozenList[JsonObj]],
                Coproduct[UnhandledErrors, MaxRetriesReached],
            ]
        ],
    ]
    post: Callable[
        [RelativeEndpoint, Token, JsonObj],
        Cmd[
            Result[
                Coproduct[JsonObj, FrozenList[JsonObj]],
                Coproduct[UnhandledErrors, MaxRetriesReached],
            ]
        ],
    ]

    get_response: Callable[
        [RelativeEndpoint, Token, JsonObj],
        Cmd[Result[Response, Coproduct[UnhandledErrors, MaxRetriesReached]]],
    ]



import hashlib
from dataclasses import (
    dataclass,
)

ACCOUNTS_URL = "https://accounts.zoho.com"  # for US region


@dataclass(frozen=True)
class Token:
    raw_token: str

    def __repr__(self) -> str:
        return "[masked]"


@dataclass(frozen=True)
class RefreshToken:
    raw_token: str

    def __repr__(self) -> str:
        signature = hashlib.sha256(self.raw_token.encode("utf-8")).hexdigest()[:128]
        return f"[masked] signature={signature}"


@dataclass(frozen=True)
class Credentials:
    client_id: str
    client_secret: str
    refresh_token: str
    scopes: frozenset[str]

    def __repr__(self) -> str:
        return f"Creds(client_id={self.client_id}, scopes={self.scopes})"


__all__ = [
    "Token",
]



import logging
import tempfile
from typing import IO
from zipfile import ZipFile

from fa_purity import Cmd, Result, ResultE, cast_exception
from pure_requests.retry import sleep_cmd
from requests import Response

from .core import (
    BulkData,
    BulkExportId,
    BulkStatus,
    FileName,
    MustRestart,
    NonTerminalStatus,
    ShouldRetry,
)

LOG = logging.getLogger(__name__)


def _top_truncation(value: float, limit: float) -> float:
    if value > limit:
        return limit
    return value


def _waiting_msg(job: BulkExportId, time: float) -> Cmd[None]:
    return Cmd.wrap_impure(
        lambda: LOG.info("Waiting bulk export %s to be ready (%s)", job.id_bulk, int(time)),
    )


def _wait_job(retry: int, job: BulkExportId) -> Cmd[None]:
    wait_time = _top_truncation(60 * retry, 5 * 50)
    return _waiting_msg(job, wait_time) + sleep_cmd(wait_time)


def _handle_status(status_bulk: BulkStatus) -> ResultE[BulkStatus]:
    if status_bulk == BulkStatus.COMPLETED:
        return Result.success(status_bulk)
    if status_bulk in NonTerminalStatus:
        return Result.failure(ShouldRetry(status_bulk))
    return Result.failure(MustRestart(status_bulk))


def unzip_data(response: Response, file_name: FileName) -> Cmd[ResultE[BulkData]]:
    def _action() -> ResultE[BulkData]:
        # pylint: disable=consider-using-with
        # need refac of BulkData for enabling the above check
        name_file = file_name.name
        tmp_zipdir = tempfile.mkdtemp()
        file_zip: IO[bytes] = tempfile.NamedTemporaryFile(mode="wb+")  # noqa: SIM115
        file_unzip: IO[str] = tempfile.NamedTemporaryFile(mode="w+", encoding="utf-8")  # noqa: SIM115
        file_zip.write(response.content)
        file_zip.seek(0)
        LOG.debug("Unzipping file")
        with ZipFile(file_zip, "r") as zip_obj:
            files = zip_obj.namelist()

            if name_file not in files:
                err = ValueError(f"Expected  {name_file} file. Decompressed {len(files)} files.")
                return Result.failure(cast_exception(err))
            zip_obj.extract(name_file, tmp_zipdir)
        LOG.debug("Generating BulkData")
        with open(tmp_zipdir + f"/{name_file}", encoding="UTF-8") as unzipped:  # noqa: PTH123
            file_unzip.write(unzipped.read())
        LOG.debug("Unzipped size: %s", file_unzip.tell())
        return Result.success(BulkData(file_unzip))

    return Cmd.wrap_impure(_action)



from collections.abc import Callable
from typing import (
    IO,
    TypeVar,
)

from fa_purity import (
    Coproduct,
    FrozenList,
    Maybe,
    PureIterFactory,
    Result,
    ResultE,
    ResultTransform,
    Unsafe,
)
from fa_purity.date_time import DatetimeUTC
from fa_purity.json import (
    JsonObj,
    JsonPrimitiveUnfolder,
    JsonUnfolder,
    UnfoldedFactory,
    Unfolder,
)
from fluidattacks_etl_utils.decode import DecodeUtils
from fluidattacks_etl_utils.natural import Natural

from fluidattacks_zoho_sdk.auth import Credentials
from fluidattacks_zoho_sdk.ids import (
    AccountId,
    ContactId,
    CrmId,
    DeparmentId,
    ProductId,
    ProfileId,
    RoleId,
    TeamId,
    TicketId,
    UserId,
)

_T = TypeVar("_T")


def decode_list_objs(
    items: FrozenList[JsonObj],
    transaform: Callable[[JsonObj], ResultE[_T]],
) -> ResultE[FrozenList[_T]]:
    if not items:
        return Result.failure(ValueError("Expected a list"))

    return ResultTransform.all_ok(
        PureIterFactory.from_list(items).map(lambda v: transaform(v)).to_list(),
    )


def _decode_zoho_creds(raw: JsonObj) -> ResultE[Credentials]:
    client_id = JsonUnfolder.require(raw, "client_id", Unfolder.to_primitive).bind(
        JsonPrimitiveUnfolder.to_str,
    )
    client_secret = JsonUnfolder.require(raw, "client_secret", Unfolder.to_primitive).bind(
        JsonPrimitiveUnfolder.to_str,
    )
    refresh_token = JsonUnfolder.require(raw, "refresh_token", Unfolder.to_primitive).bind(
        JsonPrimitiveUnfolder.to_str,
    )
    scopes_result = JsonUnfolder.require(
        raw,
        "scopes",
        lambda i: Unfolder.to_list_of(
            i,
            lambda x: Unfolder.to_primitive(x).bind(JsonPrimitiveUnfolder.to_str),
        ),
    )
    return client_id.bind(
        lambda cid: client_secret.bind(
            lambda secret: refresh_token.bind(
                lambda token: scopes_result.map(
                    lambda scopes: Credentials(cid, secret, token, frozenset(scopes)),
                ),
            ),
        ),
    )


def get_sub_json(raw: JsonObj, key: str) -> JsonObj:
    return (
        JsonUnfolder.require(raw, key, lambda v: Unfolder.to_json(v))
        .alt(Unsafe.raise_exception)
        .to_union()
    )


def decode_optional_date(raw: JsonObj, key: str) -> ResultE[Maybe[DatetimeUTC]]:
    return JsonUnfolder.optional(raw, key, DecodeUtils.to_opt_date_time).map(
        lambda v: v.bind(lambda j: j),
    )


def decode_require_date(raw: JsonObj, key: str) -> ResultE[DatetimeUTC]:
    return JsonUnfolder.require(raw, key, DecodeUtils.to_date_time)


def decode_maybe_str(raw: JsonObj, key: str) -> ResultE[Maybe[str]]:
    return JsonUnfolder.optional(raw, key, DecodeUtils.to_opt_str).map(
        lambda v: v.bind(lambda j: j),
    )


def decode_zoho_creds(auth_file: IO[str]) -> ResultE[Credentials]:
    return UnfoldedFactory.load(auth_file).bind(_decode_zoho_creds)


def decode_user_id(raw: JsonObj) -> ResultE[UserId]:
    return (
        JsonUnfolder.require(raw, "id", DecodeUtils.to_str)
        .bind(lambda v: Natural.from_int(int(v)))
        .map(UserId)
    )


def decode_rol_id(raw: JsonObj) -> ResultE[RoleId]:
    return (
        JsonUnfolder.require(raw, "roleId", DecodeUtils.to_str)
        .bind(lambda v: Natural.from_int(int(v)))
        .map(RoleId)
    )


def decode_profile_id(raw: JsonObj) -> ResultE[ProfileId]:
    return (
        JsonUnfolder.require(raw, "profileId", DecodeUtils.to_str)
        .bind(lambda v: Natural.from_int(int(v)))
        .map(ProfileId)
    )


def decode_account_id(raw: JsonObj) -> ResultE[AccountId]:
    return (
        JsonUnfolder.require(raw, "accountId", DecodeUtils.to_str)
        .bind(lambda v: Natural.from_int(int(v)))
        .map(AccountId)
    )


def decode_crm_id(raw: JsonObj) -> ResultE[CrmId]:
    return (
        JsonUnfolder.require(get_sub_json(raw, "zohoCRMContact"), "id", DecodeUtils.to_str)
        .bind(lambda v: Natural.from_int(int(v)))
        .map(lambda obj: CrmId(obj))
    )


def decode_deparment_id(raw: JsonObj) -> ResultE[DeparmentId]:
    return (
        JsonUnfolder.optional(raw, "departmentId", DecodeUtils.to_opt_str)
        .map(
            lambda v: v.bind(lambda x: x).map(
                lambda j: Natural.from_int(int(j)).alt(Unsafe.raise_exception).to_union(),
            ),
        )
        .map(lambda obj: DeparmentId(obj))
    )


def decode_product_id(raw: JsonObj) -> ResultE[ProductId]:
    return (
        JsonUnfolder.optional(raw, "productId", DecodeUtils.to_opt_str)
        .map(
            lambda v: v.bind(lambda x: x).map(
                lambda j: Natural.from_int(int(j)).alt(Unsafe.raise_exception).to_union(),
            ),
        )
        .map(lambda obj: ProductId(obj))
    )


def decode_team_id(raw: JsonObj) -> ResultE[TeamId]:
    return JsonUnfolder.require(raw, "teamId", DecodeUtils.to_str).bind(
        lambda v: Natural.from_int(int(v)).map(lambda obj: TeamId(obj)),
    )


def decode_ticket_id(raw: JsonObj) -> ResultE[TicketId]:
    return JsonUnfolder.require(raw, "id", DecodeUtils.to_str).bind(
        lambda v: Natural.from_int(int(v)).map(lambda obj: TicketId(obj)),
    )


def decode_contact_id(raw: JsonObj) -> ResultE[ContactId]:
    return JsonUnfolder.require(get_sub_json(raw, "contact"), "id", DecodeUtils.to_str).bind(
        lambda v: Natural.from_int(int(v)).map(lambda obj: ContactId(obj)),
    )


def decode_account_id_ticket(raw: JsonObj) -> ResultE[AccountId]:
    account_obj = get_sub_json(raw, "contact")
    return JsonUnfolder.require(
        get_sub_json(account_obj, "account"),
        "id",
        DecodeUtils.to_str,
    ).bind(lambda v: Natural.from_int(int(v)).map(lambda obj: AccountId(obj)))


def decode_id_team(raw: JsonObj) -> ResultE[TeamId]:
    return JsonUnfolder.require(raw, "id", DecodeUtils.to_str).bind(
        lambda v: Natural.from_int(int(v)).map(lambda obj: TeamId(obj)),
    )


def decode_user_id_bulk(raw: JsonObj) -> ResultE[UserId]:
    return JsonUnfolder.require(raw, "ID", DecodeUtils.to_str).bind(
        lambda v: Natural.from_int(int(v)).map(UserId),
    )


def assert_single(item: Coproduct[JsonObj, FrozenList[JsonObj]]) -> ResultE[JsonObj]:
    return item.map(
        Result.success,
        lambda _: Result.failure(ValueError("Expected a json not a list")),
    )


def assert_multiple(item: Coproduct[JsonObj, FrozenList[JsonObj]]) -> ResultE[FrozenList[JsonObj]]:
    return item.map(
        lambda _: Result.failure(ValueError("Expected a json list not a single json")),
        Result.success,
    )



from __future__ import annotations

from dataclasses import dataclass

from fluidattacks_zoho_sdk._http_client import ClientFactory, HttpJsonClient
from fluidattacks_zoho_sdk.auth import Credentials
from fluidattacks_zoho_sdk.ids import OrgId

from . import _client
from .core import BulkClient, BulkEndpoint, FileName, ModuleName


def _from_client(client: HttpJsonClient) -> BulkClient:
    return BulkClient(
        lambda t, m, i: _client.create_bulk_export(client, t, m, i),
        lambda token, obj: _client.get_status_bulk_export(client, token, obj),
        lambda t, b, f: _client.download_bulk(client, t, b, f),
        lambda t, m, e, f: _client.fetch_bulk(client, t, m, e, f),
    )


@dataclass(frozen=True)
class BulkApiFactory:
    @staticmethod
    def new(creds: Credentials, org_id: OrgId) -> BulkClient:
        return _from_client(ClientFactory.new(creds, org_id))


__all__ = ["BulkApiFactory", "BulkClient", "BulkEndpoint", "FileName", "ModuleName"]



from collections.abc import Callable
from typing import NewType, TypeVar

from fa_purity import Cmd, FrozenList, Maybe, Result, ResultE
from fa_purity._core.utils import raise_exception

from fluidattacks_zoho_sdk.auth import Token

from ._core import HttpJsonClient

_T = TypeVar("_T")
FromIndex = NewType("FromIndex", int)
Limit = NewType("Limit", int)


def validate_next_page(
    from_index: FromIndex,
    items: FrozenList[_T],
    limit: Limit,
) -> ResultE[tuple[FrozenList[_T], Maybe[FromIndex]]]:
    return Result.success(
        (items, Maybe.some(FromIndex(from_index + limit)) if len(items) > 0 else Maybe.empty()),
    )


def get_page(
    client: HttpJsonClient,
    token: Token,
    from_index: Maybe[FromIndex],
    limit: Limit,
    get_endpoint: Callable[
        [HttpJsonClient, Token, Maybe[FromIndex], Limit],
        Cmd[ResultE[tuple[FrozenList[_T], Maybe[FromIndex]]]],
    ],
) -> Cmd[tuple[FrozenList[_T], Maybe[FromIndex]]]:
    return get_endpoint(client, token, from_index, limit).map(
        lambda r: r.alt(raise_exception).to_union(),
    )



import inspect
import logging

from fa_purity import (
    Cmd,
    FrozenDict,
    ResultE,
    cast_exception,
)
from fa_purity.json import (
    JsonObj,
    JsonPrimitiveUnfolder,
    JsonUnfolder,
    Primitive,
    UnfoldedFactory,
    Unfolder,
)
from fluidattacks_etl_utils.bug import Bug
from pure_requests.basic import (
    Data,
    Endpoint,
    HttpClientFactory,
    Params,
)

from ._core import (
    Credentials,
    Token,
)
from ._decode import (
    decode_json,
)

ACCOUNTS_URL = "https://accounts.zoho.com"  # for US region
LOG = logging.getLogger(__name__)


def _decode(raw: JsonObj) -> ResultE[Token]:
    return JsonUnfolder.require(
        raw,
        "access_token",
        lambda v: Unfolder.to_primitive(v).bind(JsonPrimitiveUnfolder.to_str),
    ).map(Token)


def new_access_token(credentials: Credentials) -> Cmd[Token]:
    LOG.info("Generating access token")
    endpoint = Endpoint(f"{ACCOUNTS_URL}/oauth/v2/token")
    params: dict[str, Primitive] = {
        "refresh_token": credentials.refresh_token,
        "client_id": credentials.client_id,
        "client_secret": credentials.client_secret,
        "grant_type": "refresh_token",
    }
    empty: JsonObj = FrozenDict({})
    client = HttpClientFactory.new_client(None, None, None)
    response = client.post(endpoint, Params(UnfoldedFactory.from_dict(params)), Data(empty)).map(
        lambda r: Bug.assume_success(
            "NewTokenError.response",
            inspect.currentframe(),
            (str(client), endpoint.raw, str(params)),
            r.alt(cast_exception),
        ),
    )
    data = response.map(
        lambda r: Bug.assume_success(
            "NewTokenError.decode_json",
            inspect.currentframe(),
            (str(client), endpoint.raw, str(params)),
            decode_json(r).alt(
                lambda c: c.map(
                    cast_exception,
                    lambda v: v.map(cast_exception, cast_exception),
                ),
            ),
        ),
    )

    return data.map(
        lambda j: Bug.assume_success(
            "NewTokenError.decode_json",
            inspect.currentframe(),
            (JsonUnfolder.dumps(j),),
            _decode(j),
        ),
    )



import inspect
import logging
from getpass import (
    getpass,
)
from urllib.parse import (
    urlencode,
)

from fa_purity import (
    Cmd,
    FrozenDict,
    Result,
    ResultE,
    cast_exception,
)
from fa_purity.json import (
    JsonObj,
    JsonPrimitiveUnfolder,
    JsonUnfolder,
    Primitive,
    UnfoldedFactory,
    Unfolder,
)
from fluidattacks_etl_utils.bug import Bug
from pure_requests.basic import (
    Data,
    Endpoint,
    HttpClientFactory,
    Params,
)
from requests import (
    RequestException,
    Response,
)

from ._core import (
    ACCOUNTS_URL,
    Credentials,
    RefreshToken,
)
from ._decode import (
    decode_json,
)

LOG = logging.getLogger(__name__)


def _decode(raw: JsonObj) -> ResultE[RefreshToken]:
    LOG.debug("raw token response: %s", JsonUnfolder.dumps(raw))
    return JsonUnfolder.require(
        raw,
        "refresh_token",
        lambda v: Unfolder.to_primitive(v).bind(JsonPrimitiveUnfolder.to_str),
    ).map(RefreshToken)


def _user_input_code() -> Cmd[str]:
    def _action() -> str:
        LOG.info("Paste grant token:")
        token = getpass()
        LOG.debug(token)
        return token

    return Cmd.wrap_impure(_action)


def _gen_refresh_token(
    credentials: Credentials,
    code: str,
) -> Cmd[Result[Response, RequestException]]:
    endpoint = Endpoint(f"{ACCOUNTS_URL}/oauth/v2/token")
    raw: dict[str, Primitive] = {
        "grant_type": "authorization_code",
        "client_id": credentials.client_id,
        "client_secret": credentials.client_secret,
        "code": code,
    }
    empty: JsonObj = FrozenDict({})
    encoded = urlencode(raw)
    headers: dict[str, Primitive] = {
        "Content-Type": "application/x-www-form-urlencoded",
        "Content-Length": str(len(encoded)),
    }
    client = HttpClientFactory.new_client(None, UnfoldedFactory.from_dict(headers), None)
    return client.post(endpoint, Params(empty), Data(encoded))


def _decode_refresh(result: Result[Response, RequestException]) -> ResultE[RefreshToken]:
    response = Bug.assume_success(
        "AuthRefreshError.response",
        inspect.currentframe(),
        (),
        result.alt(cast_exception),
    )
    json_data = Bug.assume_success(
        "AuthRefreshError.decode_json",
        inspect.currentframe(),
        (),
        decode_json(response).alt(
            lambda e: e.map(cast_exception, lambda f: f.map(cast_exception, cast_exception)),
        ),
    )
    return _decode(json_data)


def generate_refresh_token(
    credentials: Credentials,
) -> Cmd[RefreshToken]:
    msg = Cmd.wrap_impure(
        lambda: LOG.info(
            "Generating refresh token with scopes: %s",
            ",".join(credentials.scopes),
        ),
    )
    return msg + _user_input_code().bind(lambda c: _gen_refresh_token(credentials, c)).map(
        lambda r: Bug.assume_success(
            "AuthRefreshError.decode_token",
            inspect.currentframe(),
            (str(r),),
            _decode_refresh(r),
        ),
    )



from __future__ import annotations

from collections.abc import Callable
from dataclasses import dataclass

from fa_purity import Cmd, FrozenList, Maybe, ResultE, Stream
from fa_purity.date_time import DatetimeUTC

from fluidattacks_zoho_sdk._http_client.paginate import Limit
from fluidattacks_zoho_sdk.auth import Token
from fluidattacks_zoho_sdk.ids import (
    AccountId,
    ContactId,
    CrmId,
    DeparmentId,
    ProductId,
    TeamId,
    TicketId,
    UserId,
)


@dataclass(frozen=True)
class UserObj:
    id_user: UserId
    first_name: Maybe[str]
    last_name: str


@dataclass(frozen=True)
class AgentObj:
    user: UserObj
    is_confirmed: bool
    email: str
    profile: str
    role: str
    status: str


@dataclass(frozen=True)
class ContactDates:
    created_time: DatetimeUTC
    modified_time: DatetimeUTC


@dataclass(frozen=True)
class ContactInfo:
    email: str
    facebook: Maybe[str]
    phone: str
    mobile: Maybe[str]
    secondary_email: Maybe[str]


@dataclass(frozen=True)
class ContactAddres:
    city: Maybe[str]
    country: Maybe[str]
    street: Maybe[str]


@dataclass(frozen=True)
class ContactObj:
    user: UserObj
    contact_info: ContactInfo
    contact_addres: ContactAddres
    contact_dates: ContactDates
    account_id: AccountId
    contact_owner: UserId
    crm_id: CrmId
    description: Maybe[str]
    state: Maybe[str]
    title: Maybe[str]
    type_contact: Maybe[str]
    zip_contact: Maybe[str]


@dataclass(frozen=True)
class TeamObj:
    id_deparment: DeparmentId
    id_team: TeamId
    description: str
    name: str


@dataclass(frozen=True)
class TicketDates:
    modified_time: Maybe[DatetimeUTC]
    created_time: DatetimeUTC
    closed_time: Maybe[DatetimeUTC]
    customer_response_time: Maybe[DatetimeUTC]


@dataclass(frozen=True)
class TicketProperties:
    id_ticket: TicketId
    ticket_number: str
    suject: Maybe[str]
    channel: Maybe[str]
    status: Maybe[str]
    category: Maybe[str]
    is_escalated: bool
    priority: str
    resolution: Maybe[str]
    classification: Maybe[str]
    description: Maybe[str]


@dataclass(frozen=True)
class TicketObj:
    account_id: AccountId
    deparment_id: DeparmentId
    team_id: TeamId
    product_id: ProductId
    ticket_date: TicketDates
    ticket_properties: TicketProperties
    contact_id: ContactId
    email: str
    phone: str
    onhold_time: Maybe[str]


@dataclass(frozen=True)
class DeskClient:
    get_contacts: Callable[[Token, Limit], Cmd[Stream[FrozenList[ContactObj]]]]
    get_tickets: Callable[[Token, Limit], Cmd[Stream[FrozenList[TicketObj]]]]
    get_teams: Callable[[Token], Cmd[ResultE[FrozenList[TeamObj]]]]
    # bulk export
    fetch_bulk_agents: Callable[[Token], Cmd[ResultE[FrozenList[AgentObj]]]]



from arch_lint.dag import (
    DagMap,
)
from arch_lint.graph import (
    FullPathModule,
)
from fluidattacks_etl_utils.typing import (
    Dict,
    FrozenSet,
    NoReturn,
    TypeVar,
)

_T = TypeVar("_T")


def raise_or_return(item: _T | Exception) -> _T | NoReturn:
    if isinstance(item, Exception):
        raise item
    return item


def _module(path: str) -> FullPathModule | NoReturn:
    return raise_or_return(FullPathModule.from_raw(path))


_dag: Dict[str, tuple[tuple[str, ...] | str, ...]] = {
    "fluidattacks_zoho_sdk": (
        ("zoho_desk"),
        ("auth", "_http_client", "bulk_export", "_decoders"),
        "ids",
    ),
    "fluidattacks_zoho_sdk.auth": (
        ("_access", "_refresh", "_revoke"),
        "_decode",
        "_core",
    ),
    "fluidattacks_zoho_sdk._http_client": (
        "_client",
        "paginate",
        "_core",
    ),
    "fluidattacks_zoho_sdk.bulk_export": (
        "_client",
        "_decode",
        "utils",
        "core",
    ),
    "fluidattacks_zoho_sdk.zoho_desk": (
        "_client",
        "_decode",
        "core",
    ),
}


def project_dag() -> DagMap:
    return raise_or_return(DagMap.new(_dag))


def forbidden_allowlist() -> Dict[FullPathModule, FrozenSet[FullPathModule]]:
    _raw: Dict[str, FrozenSet[str]] = {}
    return {_module(k): frozenset(_module(i) for i in v) for k, v in _raw.items()}



from dataclasses import dataclass

from fluidattacks_zoho_sdk.auth import Credentials
from fluidattacks_zoho_sdk.ids import OrgId

from ._client import Client
from ._core import HttpJsonClient, RelativeEndpoint


@dataclass(frozen=True)
class ClientFactory:
    @staticmethod
    def new(creds: Credentials, org_id: OrgId) -> HttpJsonClient:
        return Client.new(creds, org_id).client


__all__ = ["Client", "HttpJsonClient", "RelativeEndpoint"]



import csv

from fa_purity import FrozenList, ResultE
from fa_purity.json import JsonObj, JsonUnfolder, JsonValueFactory, Unfolder
from fluidattacks_etl_utils import smash
from fluidattacks_etl_utils.decode import DecodeUtils

from .core import BulkData, BulkExportId, BulkExportObj, BulkStatus, ViewId


def decode_id_bulk(raw: JsonObj) -> ResultE[BulkExportId]:
    return JsonUnfolder.require(raw, "exportId", DecodeUtils.to_str).map(lambda v: BulkExportId(v))


def decode_id_view(raw: JsonObj) -> ResultE[ViewId]:
    return JsonUnfolder.require(raw, "viewId", DecodeUtils.to_str).map(lambda v: ViewId(v))


def decode_status_bulk(raw: JsonObj) -> ResultE[BulkStatus]:
    return JsonUnfolder.require(raw, "status", DecodeUtils.to_str).bind(
        lambda v: BulkStatus.from_raw(v),
    )


def decode_bulk_export(raw: JsonObj) -> ResultE[BulkExportObj]:
    return smash.smash_result_2(
        decode_id_bulk(raw),
        JsonUnfolder.require(raw, "module", DecodeUtils.to_str),
    ).map(lambda bulk: BulkExportObj(*bulk))


def decode_bulk_info(raw: JsonObj) -> ResultE[tuple[BulkExportObj, BulkStatus]]:
    return smash.smash_result_2(decode_bulk_export(raw), decode_status_bulk(raw))


def transform_columns(bulk: BulkData) -> ResultE[FrozenList[JsonObj]]:
    bulk.file.seek(0)
    reader = csv.DictReader(bulk.file)
    rows: list[dict[str, str | None]] = list(reader)

    return JsonValueFactory.from_any(rows).bind(lambda v: Unfolder.to_list_of(v, Unfolder.to_json))



from datetime import UTC, datetime
from pathlib import Path

from fa_purity import Maybe, Unsafe
from fa_purity.date_time import DatetimeUTC
from fa_purity.json import JsonObj, JsonValueFactory, Unfolder
from fluidattacks_etl_utils.natural import NaturalOperations

from fluidattacks_zoho_sdk.ids import (
    AccountId,
    ContactId,
    CrmId,
    DeparmentId,
    ProductId,
    TeamId,
    TicketId,
    UserId,
)
from fluidattacks_zoho_sdk.zoho_desk._decode import (
    decode_agent_obj,
    decode_contact_obj,
    decode_team_obj,
    decode_ticket_obj,
)
from fluidattacks_zoho_sdk.zoho_desk.core import (
    AgentObj,
    ContactAddres,
    ContactDates,
    ContactInfo,
    ContactObj,
    TeamObj,
    TicketDates,
    TicketObj,
    TicketProperties,
    UserObj,
)


def read_json(name_file: str) -> JsonObj:
    raw_data_contact = Path(__file__).parent / name_file
    return (
        JsonValueFactory.load(raw_data_contact.open(encoding="utf=8"))
        .bind(Unfolder.to_json)
        .alt(Unsafe.raise_exception)
        .to_union()
    )


def _assert_utc(date_time: datetime) -> DatetimeUTC:
    return DatetimeUTC.assert_utc(date_time).alt(Unsafe.raise_exception).to_union()


def test_decode_agent() -> None:
    agent_expected = AgentObj(
        UserObj(
            UserId(NaturalOperations.absolute(1892000000056007)),
            Maybe.some("Arnol"),
            "case",
        ),
        True,
        "case@zylker.com",
        "Administrator",
        "CEO",
        "ACTIVE",
    )

    file_name = "data_agent.json"

    decode_data = decode_agent_obj(read_json(file_name)).alt(Unsafe.raise_exception).to_union()
    assert agent_expected == decode_data


def test_decode_contact() -> None:
    expected_data = ContactObj(
        UserObj(
            UserId(NaturalOperations.absolute(3263000000064001)),
            Maybe.some("Jade"),
            "Smith",
        ),
        ContactInfo(
            "lawrence@zylker.com",
            Maybe.empty(),
            "123 99 888 23",
            Maybe.empty(),
            Maybe.empty(),
        ),
        ContactAddres(
            Maybe.empty(),
            Maybe.empty(),
            Maybe.empty(),
        ),
        ContactDates(
            _assert_utc(datetime(2015, 2, 16, 14, 46, 24, tzinfo=UTC)),
            _assert_utc(datetime(2015, 3, 2, 14, 49, 32, tzinfo=UTC)),
        ),
        AccountId(NaturalOperations.absolute(123456)),
        UserId(NaturalOperations.absolute(3263000000057001)),
        CrmId(NaturalOperations.absolute(5000000014010)),
        Maybe.empty(),
        Maybe.empty(),
        Maybe.empty(),
        Maybe.empty(),
        Maybe.empty(),
    )
    path_file = "data_contact.json"
    decode_data = decode_contact_obj(read_json(path_file)).alt(Unsafe.raise_exception).to_union()
    assert expected_data == decode_data


def test_decode_ticket() -> None:
    expected_data = TicketObj(
        AccountId(NaturalOperations.absolute(1892000000975382)),
        DeparmentId(Maybe.some(NaturalOperations.absolute(1892000000006907))),
        TeamId(NaturalOperations.absolute(8920000000069071)),
        ProductId(Maybe.empty()),
        TicketDates(
            Maybe.some(_assert_utc(datetime(2016, 6, 21, 13, 16, 14, 0, tzinfo=UTC))),
            _assert_utc(datetime(2013, 11, 4, 11, 21, 7, 0, tzinfo=UTC)),
            Maybe.empty(),
            Maybe.some(_assert_utc(datetime(2013, 11, 4, 11, 21, 7, 912000, tzinfo=UTC))),
        ),
        TicketProperties(
            TicketId(NaturalOperations.absolute(1892000000042034)),
            "101",
            Maybe.some(
                "Hi. There is a sudden delay in the processing of the orders. Check this with high priority",  # noqa: E501
            ),
            Maybe.some("Email"),
            Maybe.some("Open"),
            Maybe.some("general"),
            False,
            "High",
            Maybe.empty(),
            Maybe.empty(),
            Maybe.some(
                "Hi. There is a sudden delay in the processing of the orders. Check this with high priority",  # noqa: E501
            ),
        ),
        ContactId(NaturalOperations.absolute(1892000000042032)),
        "carol@zylker.com",
        "1 888 900 9646",
        Maybe.empty(),
    )
    path_file = "data_ticket.json"
    decode_data = decode_ticket_obj(read_json(path_file)).alt(Unsafe.raise_exception).to_union()
    assert expected_data == decode_data


def test_decode_teams() -> None:
    expected_data = TeamObj(
        DeparmentId(Maybe.some(NaturalOperations.absolute(17000000007253))),
        TeamId(NaturalOperations.absolute(17000000013003)),
        "Sales teams for winning customers.",
        "Sales Reps.",
    )
    path_file = "data_team.json"
    decode_data = decode_team_obj(read_json(path_file)).alt(Unsafe.raise_exception).to_union()
    assert expected_data == decode_data



from typing import (
    TypeVar,
)

from fa_purity import (
    Coproduct,
    CoproductFactory,
    FrozenList,
    Result,
    ResultE,
    ResultFactory,
)
from fa_purity.json import (
    JsonObj,
)
from pure_requests.response import (
    handle_status,
    json_decode,
)
from requests import (
    HTTPError,
    JSONDecodeError,
    Response,
)

_T = TypeVar("_T")


def decode_json(
    response: Response,
) -> Result[JsonObj, Coproduct[HTTPError, Coproduct[JSONDecodeError, TypeError]]]:
    _factory: CoproductFactory[HTTPError, Coproduct[JSONDecodeError, TypeError]] = (
        CoproductFactory()
    )
    _factory_2: CoproductFactory[JSONDecodeError, TypeError] = CoproductFactory()
    _factory_3: ResultFactory[
        JsonObj,
        Coproduct[HTTPError, Coproduct[JSONDecodeError, TypeError]],
    ] = ResultFactory()
    return (
        handle_status(response)
        .alt(lambda x: _factory.inl(x))
        .bind(
            lambda r: json_decode(r)
            .alt(lambda e: _factory.inr(_factory_2.inl(e)))
            .bind(
                lambda c: c.map(
                    lambda j: _factory_3.success(j),
                    lambda _: _factory_3.failure(
                        _factory.inr(_factory_2.inr(TypeError("Expected non-list"))),
                    ),
                ),
            ),
        )
    )


def require_index(items: FrozenList[_T], index: int) -> ResultE[_T]:
    try:
        return Result.success(items[index])
    except IndexError as err:
        return Result.failure(Exception(err))



from dataclasses import dataclass

from fa_purity import Maybe
from fluidattacks_etl_utils.natural import Natural


@dataclass(frozen=True)
class OrgId:
    org_id: Natural


@dataclass(frozen=True)
class UserId:
    user_id: Natural


@dataclass(frozen=True)
class AccountId:
    accound_id: Natural


@dataclass(frozen=True)
class CrmId:
    crm_id: Natural


@dataclass(frozen=True)
class DeparmentId:
    id_deparment: Maybe[Natural]


@dataclass(frozen=True)
class TeamId:
    id_team: Natural


@dataclass(frozen=True)
class AgentId:
    agent_id: Natural


@dataclass(frozen=True)
class ProfileId:
    id_profile: Natural


@dataclass(frozen=True)
class RoleId:
    id_role: Natural


@dataclass(frozen=True)
class ProductId:
    id_product: Maybe[Natural]


@dataclass(frozen=True)
class TicketId:
    id_ticket: Natural


@dataclass(frozen=True)
class ContactId:
    id_contact: Natural



from dataclasses import dataclass

from fluidattacks_zoho_sdk._http_client import ClientFactory, HttpJsonClient
from fluidattacks_zoho_sdk.auth import Credentials
from fluidattacks_zoho_sdk.bulk_export import BulkApiFactory
from fluidattacks_zoho_sdk.bulk_export.core import BulkClient
from fluidattacks_zoho_sdk.ids import OrgId

from . import _client
from .core import AgentObj, ContactObj, DeskClient, UserObj


def _bulk_client(creds: Credentials, org_id: OrgId) -> BulkClient:
    return BulkApiFactory.new(creds, org_id)


def _from_client(client: HttpJsonClient, creds: Credentials, org_id: OrgId) -> DeskClient:
    bulk_client = _bulk_client(creds, org_id)
    return DeskClient(
        lambda t, li: _client.get_contacts(client, t, li),
        lambda t, li: _client.get_tickets(client, t, li),
        lambda t: _client.get_teams(client, t),
        lambda t: _client.fetch_bulk_export_agents(bulk_client, t),
    )


@dataclass(frozen=True)
class DeskAgentClientFactory:
    @staticmethod
    def new(creds: Credentials, org_id: OrgId) -> DeskClient:
        http_client = ClientFactory.new(creds, org_id)
        return _from_client(http_client, creds, org_id)


__all__ = [
    "AgentObj",
    "ContactObj",
    "DeskAgentClient",
    "DeskAgentClientFactory",
    "UserObj",
]



import inspect
import logging
from getpass import (
    getpass,
)
from urllib.parse import (
    urlencode,
)

from fa_purity import (
    Cmd,
    FrozenDict,
    cast_exception,
)
from fa_purity.json import (
    JsonObj,
    Primitive,
    UnfoldedFactory,
)
from fluidattacks_etl_utils.bug import Bug
from pure_requests.basic import (
    Data,
    Endpoint,
    HttpClientFactory,
    Params,
)

from ._core import (
    RefreshToken,
)
from ._decode import (
    decode_json,
)

ACCOUNTS_URL = "https://accounts.zoho.com"  # for US region
LOG = logging.getLogger(__name__)


def _user_input() -> Cmd[RefreshToken]:
    def _action() -> RefreshToken:
        LOG.info("Refresh token to revoke:")
        return RefreshToken(getpass())

    return Cmd.wrap_impure(_action)


def _revoke_refresh_token(token: RefreshToken) -> Cmd[JsonObj]:
    endpoint = Endpoint(f"{ACCOUNTS_URL}/oauth/v2/token/revoke")
    empty: JsonObj = FrozenDict({})
    raw: dict[str, Primitive] = {
        "token": token.raw_token,
    }
    encoded = urlencode(raw)
    headers: dict[str, Primitive] = {
        "Content-Type": "application/x-www-form-urlencoded",
        "Content-Length": str(len(encoded)),
    }
    client = HttpClientFactory.new_client(None, UnfoldedFactory.from_dict(headers), None)
    return (
        client.post(
            endpoint,
            Params(empty),
            Data(encoded),
        )
        .map(
            lambda r: Bug.assume_success(
                "AuthRevokeError.response",
                inspect.currentframe(),
                (endpoint.raw, str(token)),
                r.alt(cast_exception),
            ),
        )
        .map(
            lambda r: Bug.assume_success(
                "AuthRevokeError.decode",
                inspect.currentframe(),
                (endpoint.raw, str(token), str(r), str(decode_json(r))),
                decode_json(r).alt(
                    lambda e: e.map(
                        cast_exception,
                        lambda x: x.map(cast_exception, cast_exception),
                    ),
                ),
            ),
        )
    )


def revoke_refresh_token() -> Cmd[JsonObj]:
    return (
        _user_input()
        .bind(_revoke_refresh_token)
        .bind(
            lambda j: Cmd.wrap_impure(lambda: LOG.debug("revoke_refresh_token: %s", j)).map(
                lambda _: j,
            ),
        )
    )



import inspect
import logging

from fa_purity import Cmd, FrozenList, Maybe, ResultE, Stream, cast_exception
from fa_purity.json import Primitive, UnfoldedFactory
from fluidattacks_etl_utils.bug import Bug
from fluidattacks_etl_utils.decode import int_to_str
from fluidattacks_etl_utils.paginate import cursor_pagination

from fluidattacks_zoho_sdk._decoders import assert_single
from fluidattacks_zoho_sdk._http_client import HttpJsonClient, RelativeEndpoint
from fluidattacks_zoho_sdk._http_client.paginate import (
    FromIndex,
    Limit,
    get_page,
    validate_next_page,
)
from fluidattacks_zoho_sdk.auth import Token
from fluidattacks_zoho_sdk.bulk_export import BulkClient, BulkEndpoint, FileName, ModuleName

from ._decode import decode_agents, decode_contacts, decode_teams, decode_tickets
from .core import (
    AgentObj,
    ContactObj,
    TeamObj,
    TicketObj,
)

LOG = logging.getLogger(__name__)

"""
    The Zoho Desk endpoint uses pagination based on `from` and `limit` parameters

    Docs:
        https://desk.zoho.com/DeskAPIDocument#Pagination
"""


def get_contacts_endpoint(
    client: HttpJsonClient,
    access_token: Token,
    start_from: Maybe[FromIndex],
    limit: Limit,
) -> Cmd[ResultE[tuple[FrozenList[ContactObj], Maybe[FromIndex]]]]:
    endpoint = RelativeEndpoint.new("contats")
    start_from_or_default = start_from.value_or(FromIndex(1))
    params: dict[str, Primitive] = {
        "from": int_to_str(start_from_or_default),
        "limit": int_to_str(limit),
    }
    return client.get(
        endpoint,
        access_token,
        UnfoldedFactory.from_dict(params),
    ).map(
        lambda result: (
            result.alt(
                lambda e: cast_exception(Bug.new("_get_contact", inspect.currentframe(), e, ())),
            )
            .bind(assert_single)
            .bind(lambda contacts: decode_contacts(contacts))
            .bind(lambda items: validate_next_page(start_from_or_default, items, limit))
        ),
    )


def get_contacts(
    client: HttpJsonClient,
    access_token: Token,
    limit: Limit,
) -> Cmd[Stream[FrozenList[ContactObj]]]:
    stream = cursor_pagination(
        lambda next_batch: get_page(client, access_token, next_batch, limit, get_contacts_endpoint),
    )
    return Cmd.wrap_value(stream)


def get_tickets_endpoint(
    client: HttpJsonClient,
    access_token: Token,
    start_from: Maybe[FromIndex],
    limit: Limit,
) -> Cmd[ResultE[tuple[FrozenList[TicketObj], Maybe[FromIndex]]]]:
    start_from_or_default = start_from.value_or(FromIndex(1))
    endpoint = RelativeEndpoint.new("tickets")
    params: dict[str, Primitive] = {
        "from": int_to_str(start_from_or_default),
        "limit": int_to_str(limit),
    }
    return client.get(
        endpoint,
        access_token,
        UnfoldedFactory.from_dict(params),
    ).map(
        lambda result: (
            result.alt(
                lambda e: cast_exception(Bug.new("_get_contact", inspect.currentframe(), e, ())),
            )
            .bind(assert_single)
            .bind(lambda tickets: decode_tickets(tickets))
            .bind(lambda items: validate_next_page(start_from_or_default, items, limit))
        ),
    )


def get_tickets(
    client: HttpJsonClient,
    token: Token,
    limit: Limit,
) -> Cmd[Stream[FrozenList[TicketObj]]]:
    stream = cursor_pagination(
        lambda next_batch: get_page(client, token, next_batch, limit, get_tickets_endpoint),
    )
    return Cmd.wrap_value(stream)


def get_teams(
    client: HttpJsonClient,
    access_token: Token,
) -> Cmd[ResultE[FrozenList[TeamObj]]]:
    endpoint = RelativeEndpoint.new("teams")
    params: dict[str, Primitive] = {}
    return client.get(
        endpoint,
        access_token,
        UnfoldedFactory.from_dict(params),
    ).map(
        lambda result: (
            result.alt(
                lambda e: cast_exception(Bug.new("_get_teams", inspect.currentframe(), e, ())),
            )
            .bind(assert_single)
            .bind(lambda teams: decode_teams(teams))
        ),
    )


def fetch_bulk_export_agents(
    client: BulkClient,
    access_token: Token,
) -> Cmd[ResultE[FrozenList[AgentObj]]]:
    endpoint_bulk = BulkEndpoint("bulkExport")
    module_name = ModuleName("agents")
    file_name = FileName("Agents__1.csv")

    return client.fetch_bulk(access_token, module_name, endpoint_bulk, file_name).map(
        lambda result: (
            result.alt(
                lambda e: cast_exception(
                    Bug.new("_generate_bulk_export_tickets", inspect.currentframe(), e, ()),
                ),
            ).bind(decode_agents)
        ),
    )
